# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import os.path
from pyspark import SparkConf
from pyspark.sql import SparkSession
import hlink.spark
from pyspark.sql.types import (
    ArrayType,
    BooleanType,
    DoubleType,
    LongType,
    StringType,
    StructField,
    StructType,
)


class SparkConnection(object):
    """Handles initialization of spark session and connection to local cluster."""

    def __init__(self, derby_dir, warehouse_dir, tmp_dir, python, db_name):
        self.derby_dir = derby_dir
        self.warehouse_dir = warehouse_dir
        self.db_name = db_name
        self.tmp_dir = tmp_dir
        self.python = python

    def spark_conf(self, executor_cores, executor_memory, cores):
        spark_package_path = os.path.dirname(hlink.spark.__file__)
        jar_path = os.path.join(
            spark_package_path, "jars", "hlink_lib-assembly-1.0.jar"
        )
        os.environ["PYSPARK_PYTHON"] = self.python
        conf = (
            SparkConf()
            .set("spark.pyspark.python", self.python)
            .set("spark.local.dir", self.tmp_dir)
            .set("spark.sql.warehouse.dir", self.warehouse_dir)
            .set(
                "spark.driver.extraJavaOptions", f"-Dderby.system.home={self.derby_dir}"
            )
            .set("spark.executorEnv.SPARK_LOCAL_DIRS", self.tmp_dir)
            .set("spark.sql.legacy.allowUntypedScalaUDF", True)
            .setAppName("linking")
            # .set("spark.executor.cores", executor_cores) \
        )
        if executor_memory:
            conf.set("spark.executor.memory", executor_memory)
            # conf.set("spark.driver.memory", executor_memory)
        if cores:
            conf.set("spark.cores.max", cores)

        if os.path.isfile(jar_path):
            conf = conf.set("spark.jars", jar_path)
        return conf

    def local(self, cores=1, executor_memory="10G"):
        """Create a local 'cluster'."""
        return self.connect(f"local[{cores}]", cores, executor_memory, cores)

    def connect(
        self, connection_string, executor_cores=None, executor_memory=None, cores=None
    ):
        conf = self.spark_conf(
            executor_cores=executor_cores, executor_memory=executor_memory, cores=cores
        )
        session = (
            SparkSession.builder.config(conf=conf)
            .enableHiveSupport()
            .master(connection_string)
            .getOrCreate()
        )
        session.sparkContext.setLogLevel("ERROR")

        if self.db_name not in [d.name for d in session.catalog.listDatabases()]:
            session.sql(f"CREATE DATABASE IF NOT EXISTS {self.db_name}")
        session.catalog.setCurrentDatabase(self.db_name)
        session.sparkContext.setCheckpointDir(str(self.tmp_dir))
        self.__register_udfs(session)
        return session

    def __register_udfs(self, session):
        session.udf.registerJavaFunction("jw", "com.isrdi.udfs.JWCompare", DoubleType())
        session.udf.registerJavaFunction(
            "jw_max", "com.isrdi.udfs.MaxJWCompare", DoubleType()
        )
        session.udf.registerJavaFunction(
            "jw_rate", "com.isrdi.udfs.JWRate", DoubleType()
        )
        session.udf.registerJavaFunction(
            "rel_jw", "com.isrdi.udfs.JWRelatedRows", DoubleType()
        )
        session.udf.registerJavaFunction(
            "extra_children", "com.isrdi.udfs.ExtraChildren", DoubleType()
        )
        session.udf.registerJavaFunction(
            "hh_compare_rate", "com.isrdi.udfs.HHCompare", DoubleType()
        )
        session.udf.registerJavaFunction(
            "has_matching_element", "com.isrdi.udfs.HasMatchingElement", BooleanType()
        )
        session.udf.registerJavaFunction(
            "parseProbVector", "com.isrdi.udfs.ParseProbabilityVector", DoubleType()
        )
        session.udf.registerJavaFunction(
            "hh_rows_get_first_value",
            "com.isrdi.udfs.HHRowsGetFirstValue",
            StructType(
                [StructField("serial", LongType()), StructField("input", StringType())]
            ),
        )
        session.udf.registerJavaFunction(
            "extract_neighbors",
            "com.isrdi.udfs.ExtractNeighbors",
            ArrayType(StringType()),
        )



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import hlink.linking.core.comparison_feature as comparison_feature_core
import hlink.linking.core.dist_table as dist_table_core

from hlink.linking.link_step import LinkStep


class LinkStepCreateComparisonFeatures(LinkStep):
    def __init__(self, task):
        super().__init__(
            task,
            "create comparison features",
            input_table_names=[
                "prepped_df_a",
                "prepped_df_b",
                f"{task.table_prefix}training_data",
            ],
            output_table_names=[f"{task.table_prefix}training_features"],
        )

    def _run(self):
        self.task.spark.sql("set spark.sql.shuffle.partitions=200")
        self.__create_training_features()

    def __create_training_features(self):
        training_conf = str(self.task.training_conf)
        table_prefix = self.task.table_prefix
        config = self.task.link_run.config

        if config[training_conf].get("use_training_data_features", False):
            return self.task.run_register_python(
                f"{table_prefix}training_features",
                lambda: self.task.spark.table(f"{table_prefix}training_data"),
                persist=True,
            )
        id_col = config["id_column"]
        dep_var = config[training_conf]["dependent_var"]
        if training_conf == "hh_training":
            hh_col = config[training_conf].get("hh_col", "serialp")
            tdl = self.task.spark.sql(
                f"""SELECT
                                    td.{id_col}_a,
                                    td.{id_col}_b,
                                    td.{dep_var},
                                    pdfa.{hh_col} as {hh_col}_a,
                                    pdfb.{hh_col} as {hh_col}_b
                                    from
                                    {table_prefix}training_data td
                                    left join
                                    prepped_df_a pdfa
                                    on pdfa.{id_col} = td.{id_col}_a
                                    left join
                                    prepped_df_b pdfb
                                    on pdfb.{id_col} = td.{id_col}_b
                                """
            )
        else:
            tdl = self.task.spark.table(f"{table_prefix}training_data").select(
                f"{id_col}_a", f"{id_col}_b", dep_var
            )
        self.task.run_register_python(f"{table_prefix}training_data_ids", lambda: tdl)
        (
            comp_features,
            advanced_comp_features,
            hh_comp_features,
            dist_features,
        ) = comparison_feature_core.get_features(
            config, config[training_conf]["independent_vars"]
        )
        t_ctx_def = {
            "comp_features": comp_features,
            "match_feature": config[training_conf]["dependent_var"],
            "id": config["id_column"],
            "potential_matches": f"{table_prefix}training_data_ids",
        }
        join_clauses, dist_tables = dist_table_core.register_dist_tables_and_create_sql(
            self.task, dist_features
        )
        t_ctx_def["distance_table"] = join_clauses
        if len(dist_tables) > 0:
            t_ctx_def["broadcast_hints"] = dist_table_core.get_broadcast_hint(
                dist_tables
            )

        comparison_feature_core.create_feature_tables(
            self.task,
            t_ctx_def,
            advanced_comp_features,
            hh_comp_features,
            config["id_column"],
            table_name=f"{table_prefix}training_features",
        )



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from ..link_task import LinkTask

from hlink.linking.model_exploration.link_step_ingest_file import LinkStepIngestFile
from hlink.linking.model_exploration.link_step_create_features import (
    LinkStepCreateFeatures,
)
from hlink.linking.model_exploration.link_step_train_test_models import (
    LinkStepTrainTestModels,
)


class HHModelExploration(LinkTask):
    def __init__(self, link_run):
        super().__init__(
            link_run,
            display_name="Household Model Exploration",
        )
        self.training_conf = "hh_training"
        self.table_prefix = "hh_model_eval_"

    def get_steps(self):
        return [
            LinkStepIngestFile(self),
            LinkStepCreateFeatures(self),
            LinkStepTrainTestModels(self),
        ]



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from pathlib import Path
import pandas as pd

from hlink.linking.preprocessing import Preprocessing
from hlink.linking.model_exploration import ModelExploration
from hlink.linking.training import Training
from hlink.linking.matching import Matching
from hlink.linking.reporting import Reporting
from hlink.linking.hh_model_exploration import HHModelExploration
from hlink.linking.hh_training import HHTraining
from hlink.linking.hh_matching import HHMatching
from hlink.linking.table import Table

table_definitions_file = Path(__file__).parent / "table_definitions.csv"

link_task_choices = {
    "preprocessing": Preprocessing,
    "training": Training,
    "matching": Matching,
    "hh_training": HHTraining,
    "hh_matching": HHMatching,
    "model_exploration": ModelExploration,
    "hh_model_exploration": HHModelExploration,
    "reporting": Reporting,
}


class LinkRun:
    """A link run, which manages link tasks, spark tables, and related settings.

    A link run has attributes for each link task in `link_task_choices`. These can
    be accessed like normal attributes with dot notation or with the `get_task()`
    method. The link run also has a `known_tables` attribute which lists tables
    commonly generated during linking and their descriptions. See the `get_table()`
    function for a way to get access to `Table` objects by passing a string name.

    The `use_preexisting_tables` and `print_sql` attributes are boolean settings flags.

    The `trained_models` dictionary is used for communication between the matching
    and training link tasks and the household matching and household training link
    tasks.
    """

    def __init__(self, spark, config, use_preexisting_tables=True, print_sql=False):
        self.spark = spark
        self.config = config
        self.use_preexisting_tables = use_preexisting_tables
        self.print_sql = print_sql

        self.trained_models = {}

        for task_name in link_task_choices:
            link_task = link_task_choices[task_name](self)
            setattr(self, task_name, link_task)

        self._init_tables()

    def _init_tables(self):
        """Initialize `self.known_tables` from the contents of `table_definitions_file`."""
        table_defs = pd.read_csv(table_definitions_file)
        tables = []

        for table_def in table_defs.itertuples():
            hide = table_def.hide != 0
            tables.append(Table(self.spark, table_def.name, table_def.desc, hide))

        self.known_tables = {table.name: table for table in tables}

    def get_task(self, task_name: str):
        """Get a link task attribute of the link run by name.

        If you have the string name of a link task and want the task itself,
        use this method instead of something like `getattr()`.

        Args:
            task_name (str): the name of the link task

        Raises:
            AttributeError: if `task_name` is not the name of a link task on the link run

        Returns:
            LinkTask: the requested link task
        """
        if task_name in link_task_choices:
            return getattr(self, task_name)
        else:
            raise AttributeError(f"LinkRun has no task named '{task_name}'")

    def get_table(self, table_name: str):
        """Get a `Table` by name.

        If the table is in `self.known_tables`, return it. Otherwise, return a new
        table. This method is infallible, so it will always return a table. The
        returned table may or may not exist in spark.

        Args:
            table_name (str): the name of the table to retrieve

        Returns:
            Table: the requested table
        """
        if table_name in self.known_tables:
            return self.known_tables[table_name]
        return Table(self.spark, table_name, "Unknown table", hide=True)

    def drop_temp_tables(self):
        """Delete all temporary spark tables."""
        all_tables = self.spark.catalog.listTables()
        temp_tables = filter((lambda table: table.tableType == "TEMPORARY"), all_tables)

        for table in temp_tables:
            print(f"Dropping {table.name}")
            self.spark.catalog.dropTempView(table.name)



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from pyspark.sql.window import Window
from pyspark.sql.functions import rank, lead


def get_threshold_ratio(training_conf, model_conf, default=1.3):
    """Gets the threshold ratio or default from the config using the correct precedence.

    Parameters
    ----------
    training_conf: dictionary
        the config dictionary containing the training conf
    model_conf: dictionary
        the config dictionary for a specific model
    default: float
        the default value to use if the threshold is missing

    Returns
    -------
    The threshold ratio.
    """
    if "threshold_ratio" in model_conf:
        return model_conf["threshold_ratio"]
    elif "threshold_ratio" in training_conf:
        return training_conf["threshold_ratio"]
    else:
        return default


def predict_using_thresholds(
    pred_df, alpha_threshold, threshold_ratio, training_conf, id_col
):
    """Adds a prediction column to the given pred_df by applying thresholds.

    Parameters
    ----------
    pred_df: DataFrame
        a Spark DataFrame of potential matches a probability column
    alpha_threshold: float
        the alpha threshold cutoff value. No record with a probability lower than this
        value will be considered for prediction = 1.
    threshold_ratio: float
        the threshold ratio cutoff value. Ratio's refer
        to the "a" record's next best probability value.
        Only used with the "drop_duplicate_with_threshold_ratio"
        configuration value.
    training_conf: dictionary
        the training config section
    id_col: string
        the id column

    Returns
    -------
    A Spark DataFrame containing the "prediction" column as well as other intermediate columns generated to create the prediction.
    """
    use_threshold_ratio = (
        training_conf.get("decision", "") == "drop_duplicate_with_threshold_ratio"
    )

    if use_threshold_ratio:
        return _apply_threshold_ratio(
            pred_df.drop("prediction"), alpha_threshold, threshold_ratio, id_col
        )
    else:
        return _apply_alpha_threshold(
            pred_df.drop("prediction"), alpha_threshold, threshold_ratio
        )


def _apply_alpha_threshold(pred_df, alpha_threshold, threshold_ratio):
    return pred_df.selectExpr(
        "*",
        f"case when probability >= {alpha_threshold} then 1 else 0 end as prediction",
    )


def _apply_threshold_ratio(df, alpha_threshold, threshold_ratio, id_col):
    """Apply a decision threshold using the ration of a match's probability to the next closest match's probability."""
    id_a = id_col + "_a"
    id_b = id_col + "_b"
    if "probability" not in df.columns:
        raise NameError(
            'In order to calculate the threshold ratio based on probabilities, you need to have a "probability" column in your data.'
        )
    else:
        windowSpec = Window.partitionBy(df[f"{id_a}"]).orderBy(
            df["probability"].desc(), df[f"{id_b}"]
        )
        prob_rank = rank().over(windowSpec)
        prob_lead = lead(df["probability"], 1).over(windowSpec)
        return (
            df.select(
                df["*"],
                prob_rank.alias("prob_rank"),
                prob_lead.alias("second_best_prob"),
            )
            .selectExpr(
                "*",
                f"""
            IF(
              second_best_prob IS NOT NULL
              AND second_best_prob >= {alpha_threshold}
              AND prob_rank == 1,
              probability / second_best_prob,
              NULL)
            as ratio
            """,
            )
            .selectExpr(
                "*",
                f"""
            CAST(
                probability >= {alpha_threshold}
            AND prob_rank == 1
            AND (ratio > {threshold_ratio} OR ratio is NULL)
            as INT) as prediction
            """,
            )
            .drop("prob_rank")
        )



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink


def test_men_only(spark, blocking_explode_conf, matching_test_input, matching, main):
    """Exclude women from potential matches."""
    table_a, table_b = matching_test_input
    table_a.createOrReplaceTempView("prepped_df_a")
    table_b.createOrReplaceTempView("prepped_df_b")

    # Get the number of men and women; the numbers in the
    # potential matching universe should be <= men - women
    prepped_a = spark.table("prepped_df_a").toPandas()
    prepped_b = spark.table("prepped_df_b").toPandas()

    men_a = len(prepped_a.query("sex == 1"))
    men_b = len(prepped_b.query("sex == 1"))
    women_a = len(prepped_a.query("sex == 2"))
    women_b = len(prepped_b.query("sex == 2"))

    unknown_a = len(prepped_a) - (men_a + women_a)
    unknown_b = len(prepped_b) - (men_b + women_b)

    # There can be unknown SEX values
    assert (unknown_a + men_a + women_a) == len(prepped_a)
    assert (unknown_b + men_b + women_b) == len(prepped_b)

    # For the test setup to be valid there must be women to
    # start with in the fixtures
    assert women_b > 0
    assert women_a > 0

    # Limit the universe to just men.
    blocking_explode_conf["potential_matches_universe"] = [{"expression": "sex == 1"}]

    matching.run_step(0)

    # The explode step will take the spark versions of these as inputs
    univ_a = spark.table("match_universe_df_a").toPandas()
    univ_b = spark.table("match_universe_df_b").toPandas()

    # The exploded step will produce these tables
    exploded_univ_a = spark.table("exploded_df_a").toPandas()
    exploded_univ_b = spark.table("exploded_df_b").toPandas()

    assert all(
        elem in list(exploded_univ_a.columns)
        for elem in ["namefrst", "namelast", "sex", "birthyr_3"]
    )
    assert all(
        elem in list(exploded_univ_b.columns)
        for elem in ["namefrst", "namelast", "sex", "birthyr_3"]
    )

    # Check there are no women in either universe.
    assert len(univ_a.query("sex == 2")) == 0
    assert len(univ_b.query("sex == 2")) == 0

    # Shows there aren't any unknown values of SEX
    assert men_a == len(univ_a)
    assert men_b == len(univ_b)

    assert men_a == len(univ_a.query("sex == 1"))
    assert men_b == len(univ_b.query("sex == 1"))

    # Check that there are no women in the exploded rows
    assert len(exploded_univ_a.query("sex == 2")) == 0
    assert len(exploded_univ_b.query("sex == 2")) == 0



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from ..link_task import LinkTask
from .link_step_block_on_households import LinkStepBlockOnHouseholds
from .link_step_filter import LinkStepFilter
from hlink.linking.matching.link_step_score import LinkStepScore


class HHMatching(LinkTask):
    def get_steps(self):
        return [
            LinkStepBlockOnHouseholds(self),
            LinkStepFilter(self),
            LinkStepScore(self),
        ]

    def __init__(self, link_run):
        super().__init__(link_run, display_name="Household Matching")
        self.training_conf = "hh_training"
        self.table_prefix = "hh_"



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from pyspark.ml import Transformer
from pyspark.ml.param.shared import HasInputCols
from pyspark.ml.util import DefaultParamsWritable, DefaultParamsReadable
from pyspark import keyword_only


class FloatCastTransformer(
    Transformer, HasInputCols, DefaultParamsReadable, DefaultParamsWritable
):
    """
    A custom Transformer which casts the input column to a float.
    """

    @keyword_only
    def __init__(self, inputCols=None):
        super(FloatCastTransformer, self).__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(self, inputCols=None):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def _transform(self, df):
        other_cols = set(df.columns) - set(self.getInputCols())
        casted_cols = [
            f"CAST({inputCol} as float) as {inputCol}"
            for inputCol in self.getInputCols()
        ]
        return df.selectExpr(list(other_cols) + casted_cols)



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import logging

import hlink.linking.core.comparison_feature as comparison_feature_core
import hlink.linking.core.dist_table as dist_table_core
import hlink.linking.core.comparison as comparison_core
from hlink.linking.util import spark_shuffle_partitions_heuristic
from . import _helpers as matching_helpers

from hlink.linking.link_step import LinkStep


class LinkStepMatch(LinkStep):
    def __init__(self, task):
        super().__init__(
            task,
            "match",
            input_table_names=["exploded_df_a", "exploded_df_b"],
            output_table_names=["potential_matches"],
        )

    def _run(self):
        config = self.task.link_run.config

        dataset_size_a = self.task.spark.table("exploded_df_a").count()
        dataset_size_b = self.task.spark.table("exploded_df_b").count()
        dataset_size_max = max(dataset_size_a, dataset_size_b)
        num_partitions = spark_shuffle_partitions_heuristic(dataset_size_max)
        self.task.spark.sql(f"set spark.sql.shuffle.partitions={num_partitions}")
        logging.info(
            f"Dataset sizes are A={dataset_size_a}, B={dataset_size_b}, so set Spark partitions to {num_partitions} for this step"
        )

        blocking = matching_helpers.get_blocking(config)

        t_ctx = {}
        if config.get("comparisons", False):
            if config["comparisons"] != {}:
                t_ctx["matching_clause"] = comparison_core.generate_comparisons(
                    config["comparisons"],
                    config["comparison_features"],
                    config["id_column"],
                )

        t_ctx["blocking_columns"] = [bc["column_name"] for bc in blocking]

        blocking_exploded_columns = [
            bc["column_name"] for bc in blocking if "explode" in bc and bc["explode"]
        ]
        t_ctx["dataset_columns"] = [
            c
            for c in self.task.spark.table("exploded_df_a").columns
            if c not in blocking_exploded_columns
        ]

        # comp_feature_names, dist_features_to_run, features_to_run = comparison_core.get_feature_specs_from_comp(
        #     config["comparisons"], config["comparison_features"]
        # )
        if config.get("comparisons", {}):
            comps = comparison_core.get_comparison_leaves(config["comparisons"])
            comp_feature_names = [c["feature_name"] for c in comps]

            t_ctx["feature_columns"] = [
                comparison_feature_core.generate_comparison_feature(
                    f, config["id_column"], include_as=True
                )
                for f in config["comparison_features"]
                if f["alias"] in comp_feature_names
            ]

            dist_feature_names = [
                c["alias"]
                for c in config["comparison_features"]
                if c["comparison_type"] in ["geo_distance"]
            ]
            dist_features_to_run = [
                c["feature_name"]
                for c in comps
                if c["feature_name"] in dist_feature_names
            ]

            if dist_features_to_run:
                dist_comps = [
                    c
                    for c in config["comparison_features"]
                    if c["alias"] in dist_features_to_run
                ]
                (
                    t_ctx["distance_table"],
                    dist_tables,
                ) = dist_table_core.register_dist_tables_and_create_sql(
                    self.task, dist_comps
                )
                if dist_tables:
                    t_ctx["broadcast_hints"] = dist_table_core.get_broadcast_hint(
                        dist_tables
                    )

        if config.get("streamline_potential_match_generation", False):
            t_ctx["dataset_columns"] = [config["id_column"]]
        try:
            self.task.run_register_sql("potential_matches", t_ctx=t_ctx, persist=True)
        finally:
            self.task.spark.sql("set spark.sql.shuffle.partitions=200")



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from jinja2 import Environment, PackageLoader, ChoiceLoader
from hlink.errors import SparkError
from timeit import default_timer as timer
import logging
from typing import Optional


class LinkTask(object):
    """Base class for link tasks.

    A link task consists of one or more `LinkStep`s and belongs to one `LinkRun`.
    The `get_steps()` function returns a list specifying which steps the task has
    and what order they should be run in.

    The `run_all_steps()` and `run_step()` functions are ways to run the link steps
    belonging to the task.

    `run_register_python()` and `run_register_sql()` are methods to be used by
    concrete subclasses for creating spark tables and performing step work.
    """

    def __init__(self, link_run, display_name: Optional[str] = None):
        self.link_run = link_run
        loader = ChoiceLoader(
            [
                PackageLoader(self.__class__.__module__),
                PackageLoader("hlink.linking", "templates/shared"),
            ]
        )
        self.jinja_env = Environment(loader=loader)

        if display_name is None:
            self.display_name = self.__class__.__name__
        else:
            self.display_name = display_name

    @property
    def spark(self):
        return self.link_run.spark

    def __str__(self):
        return self.display_name

    def get_steps(self):
        """Get a list of the steps that make up the link task.

        This abstract method must be implemented by concrete subclasses.

        Returns:
            List[LinkStep]: the link steps making up the link task
        """
        raise NotImplementedError()

    def run_all_steps(self):
        """Run all steps in order."""
        logging.info(f"Running all steps for task {self.display_name}")
        start_all = timer()
        for (i, step) in enumerate(self.get_steps()):
            print(f"Running step {i}: {step}")
            logging.info(f"Running step {i}: {step}")
            step.run()
        end_all = timer()
        elapsed_time_all = round(end_all - start_all, 2)
        print(f"Finished all in {elapsed_time_all}s")
        logging.info(f"Finished all steps in {elapsed_time_all}s")

    def run_step(self, step_num: int):
        """Run a particular step.

        Note that running steps out of order may cause errors when later steps
        depend on the work done by earlier steps and that work is not available.

        Args:
            step_num (int): the step number, used as an index into the `get_steps()` list
        """
        steps = self.get_steps()

        if step_num < 0 or step_num >= len(steps):
            steps_string = "\n\t".join(
                [f"step {i}: {steps[i].desc}" for i in range(len(steps))]
            )
            print(
                f"Error! Couldn't find step {step_num}. Valid steps are: \n\t{steps_string}"
            )
            return

        step = steps[step_num]
        step_string = f"step {step_num}: {step}"
        print(f"Running {step_string}")
        logging.info(f"Starting {step.task.display_name} - {step_string}")

        start = timer()
        step.run()
        end = timer()

        elapsed_time = round(end - start, 2)
        print(f"Finished {step_string} in {elapsed_time}s")
        logging.info(
            f"Finished {step.task.display_name} - {step_string} in {elapsed_time}s"
        )

    def run_register_python(
        self,
        name: str,
        func,
        args=[],
        persist=False,
        overwrite_preexisting_tables=False,
    ):
        """Run the given python function `func` and register the returned data
        frame with the given table `name`.
        """
        if name is not None:
            if overwrite_preexisting_tables is False:
                df = self._check_preexisting_table(name)
            else:
                df = None
            if df is None:
                df = func(*args)
                if persist:
                    df.write.mode("overwrite").saveAsTable(name)
                else:
                    df.createOrReplaceTempView(name)
                self.spark.sql(f"REFRESH TABLE {name}")
            return self.spark.table(name)
        else:
            try:
                return func(*args)
            except Exception as err:
                logging.error(err)
                raise SparkError(str(err))

    def run_register_sql(
        self,
        name: str,
        sql=None,
        template=None,
        t_ctx={},
        persist=False,
        overwrite_preexisting_tables=False,
    ):
        """Run the given sql or template (with context) and register the returned
        data frame with the given table `name`.

        Read the table from disk instead of running sql if `use_preexisting_tables`
        is set to True on the link task's `LinkRun`.

        Persist the created table if `persist` is True.
        """

        def run_sql():
            sql_to_run = self._get_sql(name, sql, template, t_ctx)
            if self.link_run.print_sql:
                print(sql_to_run)
            try:
                return self.spark.sql(sql_to_run)
            except Exception as err:
                print(f"Exception in Spark SQL: {sql_to_run}")

                logging.error(str(err))
                raise SparkError(str(err))

        return self.run_register_python(
            name=name,
            func=run_sql,
            persist=persist,
            overwrite_preexisting_tables=overwrite_preexisting_tables,
        )

    def _check_preexisting_table(self, name: str):
        table = self.link_run.get_table(name)
        if self.link_run.use_preexisting_tables and table.exists():
            print(f"Preexisting table: {name}")
            return table.df()
        return None

    def _get_sql(self, name: str, sql, template, t_ctx):
        if sql is None:
            template_file_name = template if template is not None else name
            template_path = f"{template_file_name}.sql"
            sql = self.jinja_env.get_template(f"{template_path}").render(t_ctx)
            print(f"{name} -- {template_path}")
        else:
            print(name)
        return sql



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from pyspark.sql.utils import AnalysisException
from os import path
import colorama


def print_checking(section: str):
    print(f"Checking {section}...", end=" ")


def print_ok():
    print(colorama.Fore.GREEN + "OK" + colorama.Style.RESET_ALL)


def analyze_conf(link_run):
    """Print an analysis of the configuration of the `link_run`."""
    colorama.init()

    try:
        print_checking("datasource_a")
        df_a = parse_datasource(link_run, "datasource_a")
        print_ok()

        print_checking("datasource_b")
        df_b = parse_datasource(link_run, "datasource_b")
        print_ok()

        print_checking("filters")
        check_filters(link_run.config, df_a, df_b)
        print_ok()

        print_checking("column_mappings")
        columns_available = check_column_mappings(link_run.config, df_a, df_b)
        print_ok()

        print_checking("substitution_columns")
        check_substitution_columns(link_run.config, columns_available)
        print_ok()

        print_checking("feature_selections")
        check_feature_selections(link_run.config, columns_available)
        print_ok()

        print_checking("blocking")
        check_blocking(link_run.config, columns_available)
        print_ok()

        print_checking("comparison_features")
        comp_features = check_comparison_features(link_run.config, columns_available)
        print_ok()

        print_checking("comparisons")
        check_comparisons(link_run.config, comp_features)
        print_ok()

        print_checking("pipeline_features")
        check_pipeline_features(link_run.config, comp_features)
        print_ok()

        print_checking("training")
        check_training(link_run.config, comp_features)
        print_ok()

        print_checking("hh_training")
        check_hh_training(link_run.config, comp_features)
        print_ok()
    finally:
        colorama.deinit()


def check_hh_training(config, comp_features):
    comp_features += ["jw_max_a", "jw_max_b"]
    hh_training = config.get("hh_training")
    if hh_training is None:
        return
    independent_vars = hh_training.get("independent_vars")
    if independent_vars is None:
        raise ValueError(
            "No independent_vars value specified in the [training] section."
        )
    for var in independent_vars:
        if var not in comp_features:
            raise ValueError(
                f"Within [training] the independent_var: '{var}' does not exist. Please add a specification as a [[comparison_feature]] or a [[pipeline_feature]]."
            )


def check_training(config, comp_features):
    comp_features += ["hits", "hits2", "exact_mult", "exact_all_mult"]
    training = config.get("training")
    if training is None:
        return
    independent_vars = training.get("independent_vars")
    if independent_vars is None:
        raise ValueError(
            "No independent_vars value specified in the [training] section."
        )
    for var in independent_vars:
        if var not in comp_features:
            raise ValueError(
                f"Within [training] the independent_var: '{var}' does not exist. Please add a specification as a [[comparison_feature]] or a [[pipeline_feature]]."
            )


def check_pipeline_features(config, comp_features):
    pipeline_features = config.get("pipeline_features")
    if pipeline_features is None:
        return
    for p in pipeline_features:
        input_column = p.get("input_column")
        if (input_column is not None) and (input_column not in comp_features):
            raise ValueError(
                f"Within [[pipeline_features]] the input_column: '{input_column}' is not available from a previous [[comparison_features]] or [[pipeline_features]] section. \n Available columns: \n {comp_features}"
            )
        input_columns = p.get("input_columns")
        if input_columns is not None:
            for c in input_columns:
                if c not in comp_features:
                    raise ValueError(
                        f"Within [[pipeline_features]] the input_column: '{c}' is not available from a previous [[comparison_features]] or [[pipeline_features]] section. \n Available columns: \n {comp_features}"
                    )
        output_column = p.get("output_column")
        if output_column is None:
            raise ValueError(
                "Within [[pipeline_features]] no 'output_column' specified for {p}."
            )
        comp_features.append(output_column)


def check_comparisons(config, comp_features):
    comparisons = config.get("comparisons")
    if comparisons is None:
        raise ValueError(
            "No [comparisons] section exists. Please add a [comparisons] section."
        )
    comp_a = comparisons.get("comp_a")
    comp_b = comparisons.get("comp_b")
    if comp_a is not None:
        feature_name = comp_a.get("feature_name")
        if (feature_name is not None) and (feature_name not in comp_features):
            raise ValueError(
                f"Within [comparisons] the feature_name '{feature_name}' is not available. Please add a corresponding feature in the [[comparison_features]] section. \n Available features: \n {comp_features}"
            )
    if comp_b is not None:
        feature_name = comp_b.get("feature_name")
        if (feature_name is not None) and (feature_name not in comp_features):
            raise ValueError(
                f"Within [comparisons] the feature_name '{feature_name}' is not available. Please add a corresponding feature in the [[comparison_features]] section. \n Available features: \n {comp_features}"
            )


def check_comparison_features(config, columns_available):
    comps = []
    comparison_features = config.get("comparison_features")
    if comparison_features is None:
        raise ValueError(
            "No [[comparison_features]] exist. Please add [[comparison_features]]."
        )
    for c in comparison_features:
        alias = c.get("alias")
        if alias is None:
            raise ValueError(
                f"No alias exists for a [[comparison_features]]: {c}. Please add an 'alias'."
            )
        column_name = c.get("column_name") or c.get("first_init_col")
        column_names = c.get("column_names") or c.get("mid_init_cols")
        if column_name is not None:
            if column_name not in columns_available:
                raise ValueError(
                    f"Within [[comparison_features]] the 'column_name' {column_name} is not available from a previous [[column_mappings]] or [[feature_selections]]: {c}"
                )
        if column_names is not None:
            for cname in column_names:
                if cname not in columns_available:
                    raise ValueError(
                        f"Within [[comparison_features]] the 'column_name' {cname} is not available from a previous [[column_mappings]] or [[feature_selections]]: {c}"
                    )
        comps.append(alias)
    return comps


def check_blocking(config, columns_available):
    blockings = config.get("blocking")
    if blockings is None:
        raise ValueError("No [[blocking]] exist. Please add blocking.")
    for b in blockings:
        column_name = b.get("derived_from") or b.get("column_name")
        if column_name is None:
            raise ValueError(f"Within [[blocking]] no column name is specified: {b}.")
        if column_name not in columns_available:
            raise ValueError(
                f"Within [[blocking]] the column_name of '{column_name}' is not available from an earlier [[column_mappings]] or [[feature_selections]]. \n Available columns: \n {columns_available}"
            )


def check_feature_selections(config, columns_available):
    feature_selections = config.get("feature_selections")
    if feature_selections is None:
        return
    for f in feature_selections:
        input_column = f.get("input_column")
        output_column = f.get("output_column") or f.get("output_col")
        other_col = f.get("other_col")
        if input_column is not None and input_column not in columns_available:
            raise ValueError(
                f"Within [[feature_selections]] the input_column: '{input_column}' is not created by an earlier [[column_mappings]] or [[feature_selections]]. \n Available Columns: \n {columns_available}."
            )
        if other_col is not None and other_col not in columns_available:
            raise ValueError(
                f"Within [[feature_selections]] the other_col: '{other_col}' is not created by an earlier [[column_mappings]] or [[feature_selections]]. \n Available Columns: \n {columns_available}."
            )
        if output_column is None:
            raise ValueError(
                f"No 'output_column' or 'output_col' value for [[feature_selections]]: {f}"
            )
        columns_available.append(output_column)


def check_substitution_columns(config, columns_available):
    substitution_columns = config.get("substitution_columns")
    if substitution_columns is None:
        return
    for s in substitution_columns:
        column_name = s.get("column_name")
        substitutions = s.get("substitutions")
        if column_name is None:
            raise ValueError("Within [[substitution_columns]] no 'column_name' exists.")
        if substitutions is None:
            raise ValueError(
                "Within [[substitution_columns]] no [[substitution_columns.substitutions]] exists."
            )
        for sub in substitutions:
            join_column = sub.get("join_column")
            join_value = sub.get("join_value")
            f = sub.get("substitution_file")
            if join_column is None or join_column not in columns_available:
                raise ValueError(
                    f"Within [[substitution_columns.substitutions]] the join_column '{join_column}' does not exist or is not available within columns specificed within [[column_mappings]]. \nList of available columns: \n {columns_available}"
                )
            if join_value is None:
                raise ValueError(
                    " Within [[substitution_columns.substitutions]] no 'join_value' exists."
                )
            if f is None or not path.exists(f):
                raise ValueError(
                    f" Within [[substitution_columns.substitutions]] no 'substitution_file' exists or does not point to an existing file: {f}"
                )


def check_column_mappings(config, df_a, df_b):
    column_mappings = config.get("column_mappings")
    if not column_mappings:
        raise ValueError("No [[column_mappings]] exist in the conf file.")
    columns_available = []
    for c in column_mappings:
        alias = c.get("alias")
        column_name = c.get("column_name")
        set_value_column_a = c.get("set_value_column_a")
        set_value_column_b = c.get("set_value_column_b")
        if not column_name:
            raise ValueError(
                "The following [[column_mappings]] has no 'column_name' attribute: {c}"
            )
        if set_value_column_a is None:
            if column_name.lower() not in [c.lower() for c in df_a.columns]:
                if column_name not in columns_available:
                    raise ValueError(
                        f"Within a [[column_mappings]] the column_name: '{column_name}' does not exist in datasource_a and no previous [[column_mapping]] alias exists for it. \nColumn mapping: {c}. \nAvailable columns: \n {df_a.columns}"
                    )
        if set_value_column_b is None:
            if column_name.lower() not in [c.lower() for c in df_b.columns]:
                if column_name not in columns_available:
                    raise ValueError(
                        f"Within a [[column_mappings]] the column_name: '{column_name}' does not exist in datasource_b and no previous [[column_mapping]] alias exists for it. Column mapping: {c}. Available columns: \n {df_b.columns}"
                    )
        if alias:
            columns_available.append(alias)
        else:
            columns_available.append(column_name)
    return columns_available


def check_filters(config, df_a, df_b):
    filters = config.get("filter")
    if not filters:
        return
    for f in filters:
        expression = f.get("expression")
        if not expression:
            raise ValueError("A [[filter]] has no expression value in the config.")
        try:
            df_a.where(expression)
            df_b.where(expression)
        except AnalysisException as e:
            raise ValueError(
                f"Within a [[filter]] the expression '{expression}' is not valid. Spark gives the following error: {e}."
            )


def parse_datasource(link_run, section_name: str):
    datasource = link_run.config.get(section_name)

    if not datasource:
        raise ValueError(f"Section [{section_name}] does not exist in config.")

    parquet_file = datasource.get("parquet_file")
    file = datasource.get("file")

    if not parquet_file and not file:
        raise ValueError(
            f"Within [{section_name}] neither 'parquet_file' nor 'file' exist."
        )
    if parquet_file and file:
        raise ValueError(
            f"Within [{section_name}] both 'parquet_file' and 'file' exist."
        )

    # Now we know that either file or parquet_file was provided, but not both.
    if parquet_file:
        if not path.exists(parquet_file):
            raise ValueError(
                f"Within [{section_name}] path of parquet file {parquet_file} does not exist."
            )
        return link_run.spark.read.parquet(parquet_file)
    else:
        if not path.exists(file):
            raise ValueError(
                f"Within [{section_name}] path of file {file} does not exist."
            )
        _, file_extension = path.splitext(file)
        if file_extension == ".csv":
            return link_run.spark.read.csv(file, header=True)
        elif file_extension == ".parquet":
            return link_run.spark.read.parquet(file)
        else:
            raise ValueError(
                f"Within [{section_name}] file {file} is neither a CSV file nor a parquet file."
            )



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from pyspark.ml import PipelineModel
from hlink.linking.link_step import LinkStep
from pathlib import Path


class LinkStepGetFeatureImportances(LinkStep):
    def __init__(self, task):
        super().__init__(
            task,
            "get feature importances",
            input_table_names=[
                f"{task.table_prefix}training_features",
                f"{task.table_prefix}training_vectorized",
            ],
            output_table_names=[f"{task.table_prefix}training_results"],
        )

    def _run(self):
        training_conf = str(self.task.training_conf)
        table_prefix = self.task.table_prefix
        config = self.task.link_run.config

        self.task.spark.sql("set spark.sql.shuffle.partitions=1")

        if "feature_importances" in config[training_conf]:
            if config[training_conf]["feature_importances"]:

                # retrieve the saved chosen model
                print("Loading chosen ML model...")
                model_path = Path(config["spark_tmp_dir"]) / "chosen_model"
                try:
                    plm = PipelineModel.load(str(model_path))
                except:
                    print(
                        "Model not found!  You might need to run step_2 to generate and train the chosen model if you haven't already done so."
                    )

                # make look at the features and their importances
                print("Retrieving model feature importances or coefficients...")
                try:
                    feature_imp = plm.stages[-2].coefficients
                except:
                    try:
                        feature_imp = plm.stages[-2].featureImportances
                    except:
                        print(
                            "This model doesn't contain a coefficient or feature importances parameter -- check chosen model type."
                        )
                    else:
                        label = "Feature importances"
                else:
                    feature_imp = feature_imp.round(4)
                    label = "Coefficients"

                varlist = self.task.spark.table(
                    f"{table_prefix}features_list"
                ).toPandas()
                for i in varlist["idx"]:
                    varlist.at[i, "score"] = feature_imp[i]
                varlist.sort_values("score", ascending=False, inplace=True)
                vl = self.spark.createDataFrame(varlist)
                vl.write.mode("overwrite").saveAsTable(
                    f"{table_prefix}feature_importances"
                )

                print(
                    f"{label} have been saved to the Spark table '{table_prefix}feature_importances'."
                )
                print(varlist)

            else:
                print(
                    f"'feature_importances' not set == true in '{training_conf}' section of config! Calculation of feature importances or coefficients not completed!"
                )
        else:
            print(
                f"'feature_importances' not included and set == true in '{training_conf}' section of config! Calculation of feature importances or coefficients not completed!"
            )

        self.task.spark.sql("set spark.sql.shuffle.partitions=200")



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink
from typing import List


class LinkStep:
    def __init__(
        self,
        task,
        desc: str,
        *,
        input_table_names: List[str] = [],
        output_table_names: List[str] = [],
        input_model_names: List[str] = [],
        output_model_names: List[str] = [],
    ):
        self.task = task
        self.desc = desc
        self.input_table_names = input_table_names
        self.output_table_names = output_table_names
        self.input_model_names = input_model_names
        self.output_model_names = output_model_names

    def find_missing_input_table_names(self):
        tables = map(self.task.link_run.get_table, self.input_table_names)
        missing_tables = filter((lambda table: not table.exists()), tables)
        return [table.name for table in missing_tables]

    def run(self):
        missing_table_names = self.find_missing_input_table_names()
        if len(missing_table_names) > 0:
            missing_names_str = ", ".join(missing_table_names)
            raise RuntimeError(
                f"Missing input tables required for link step '{self}': {missing_names_str}"
            )

        self._run()

    def _run(self):
        """Run the link step.

        This abstract method must be implemented by concrete subclasses. It is
        wrapped by the `run()` method, which makes some additional quick checks.
        """
        raise NotImplementedError()

    def __str__(self):
        return self.desc



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import pytest


@pytest.mark.skip(
    reason="We still want to test that these aggregate features are being created correctly, but we need to refactor this test to account for the fact that aggregate features are now being created in a different step (step 4 doesn't exist anymore and the functionality was moved in the code)."
)
def test_step_4_aggregate_features(
    spark, matching_conf, matching, potential_matches_agg_path
):
    """Test adding aggregate features (hits, hits2, exact_all_mult, etc.) to potential matches"""
    matching_conf["id_column"] = "histid"
    matching_conf["comparison_features"] = [
        {
            "alias": "namelast_jw",
            "column_name": "namelast",
            "comparison_type": "jaro_winkler",
        },
        {"alias": "exact"},
        {"alias": "exact_all"},
    ]
    matching_conf["training"] = {
        "independent_vars": [
            "namelast_jw",
            "exact",
            "exact_all",
            "hits",
            "hits2",
            "exact_mult",
            "exact_all_mult",
            "exact_all_mult2",
        ]
    }

    potential_matches = matching.spark.read.csv(
        potential_matches_agg_path, header=True, inferSchema=True
    )
    potential_matches.write.mode("overwrite").saveAsTable("potential_matches")
    matching.step_4_aggregate_features()

    pm_df = matching.spark.table("potential_matches").toPandas()

    assert pm_df.shape == (30, 21)
    assert (
        pm_df.query(
            "namelast_clean_a == 'cridlebaugh' and histid_b == '001B8A74-3795-4997-BC5B-2A07257668F9'"
        )["exact"].iloc[0]
        == 1
    )
    assert (
        pm_df.query(
            "namelast_clean_a == 'cridlebaugh' and histid_b == '001B8A74-3795-4997-BC5B-2A07257668F9'"
        )["exact_all"].iloc[0]
        == 1
    )
    assert (
        pm_df.query(
            "namelast_clean_a == 'cridlebaugh' and histid_b == '001B8A74-3795-4997-BC5B-2A07257668F9'"
        )["hits"].iloc[0]
        == 3
    )
    assert (
        pm_df.query(
            "namelast_clean_a == 'cridlebaugh' and histid_b == '001B8A74-3795-4997-BC5B-2A07257668F9'"
        )["hits2"].iloc[0]
        == 9
    )
    assert (
        pm_df.query(
            "namelast_clean_a == 'cridlebaugh' and histid_b == '001B8A74-3795-4997-BC5B-2A07257668F9'"
        )["exact_mult"].iloc[0]
        == 3
    )
    assert (
        pm_df.query(
            "namelast_clean_a == 'cridlebaugh' and histid_b == '001B8A74-3795-4997-BC5B-2A07257668F9'"
        )["exact_all_mult"].iloc[0]
        == 3
    )
    assert (
        pm_df.query(
            "namelast_clean_a == 'cridlebaugh' and histid_b == '001B8A74-3795-4997-BC5B-2A07257668F9'"
        )["exact_all_mult2"].iloc[0]
        == 9
    )



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import logging
from pyspark.sql.functions import col

import hlink.linking.core.column_mapping as column_mapping_core
import hlink.linking.core.substitutions as substitutions_core
import hlink.linking.core.transforms as transforms_core
from hlink.linking.util import spark_shuffle_partitions_heuristic

from hlink.linking.link_step import LinkStep


class LinkStepPrepDataframes(LinkStep):
    def __init__(self, task):
        super().__init__(
            task,
            "prepare dataframes",
            input_table_names=["raw_df_a", "raw_df_b"],
            output_table_names=["prepped_df_a", "prepped_df_b"],
        )

    def _run(self):
        config = self.task.link_run.config

        dataset_size_a = self.task.spark.table("raw_df_a").count()
        dataset_size_b = self.task.spark.table("raw_df_b").count()
        dataset_size_max = max(dataset_size_a, dataset_size_b)
        num_partitions = spark_shuffle_partitions_heuristic(dataset_size_max)
        self.task.spark.sql(f"set spark.sql.shuffle.partitions={num_partitions}")
        logging.info(
            f"Dataset sizes are A={dataset_size_a}, B={dataset_size_b}, so set Spark partitions to {num_partitions} for this step"
        )

        substitution_columns = config.get("substitution_columns", [])
        self.task.run_register_python(
            name="prepped_df_a",
            func=lambda: self._prep_dataframe(
                self.task.spark.table("raw_df_a"),
                config["column_mappings"],
                substitution_columns,
                config["feature_selections"],
                True,
                config["id_column"],
            ),
            persist=True,
        )
        self.task.run_register_python(
            name="prepped_df_b",
            func=lambda: self._prep_dataframe(
                self.task.spark.table("raw_df_b"),
                config["column_mappings"],
                substitution_columns,
                config["feature_selections"],
                False,
                config["id_column"],
            ),
            persist=True,
        )

        self.task.spark.sql("set spark.sql.shuffle.partitions=200")

    # Create a function to correctly map and select the columns from the data frames
    def _prep_dataframe(
        self,
        df,
        column_definitions,
        substitution_columns,
        feature_selections,
        is_a,
        id_column,
    ):
        """
        Returns a new dataframe after having selected the given columns out with appropriate
        transformations and substitutions.

        Parameters
        ----------
        df: dataframe to operate on
        column_definitions: config array of columns to pull out and transforms to apply
        substitution_columns: config array of substitutions to apply to the selected columns
        id_column: unique id column for a record

        Returns
        ----------
        New dataframe with the operations having been applied.
        """
        df_selected = df
        spark = self.task.spark
        column_selects = [col(id_column)]
        if column_definitions and isinstance(column_definitions[0], list):
            print(
                "DEPRECATION WARNING: The config value 'column_mappings' is no longer a nested (double) array and is now an array of objects. Please change your config for future releases."
            )
            flat_column_mappings = [
                item for sublist in column_definitions for item in sublist
            ]
        else:
            flat_column_mappings = column_definitions

        for column_mapping in flat_column_mappings:
            df_selected, column_selects = column_mapping_core.select_column_mapping(
                column_mapping, df_selected, is_a, column_selects
            )

        df_selected = df_selected.select(column_selects)

        df_selected = substitutions_core.generate_substitutions(
            spark, df_selected, substitution_columns
        )

        df_selected = transforms_core.generate_transforms(
            spark, df_selected, feature_selections, self.task, is_a, id_column
        )
        return df_selected



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from collections import namedtuple
from pyspark.sql.functions import concat_ws, lit, regexp_replace, split, when


def generate_substitutions(spark, df_selected, substitution_columns):
    for substitution_column in substitution_columns:
        column_name = substitution_column["column_name"]
        for substitution in substitution_column["substitutions"]:
            if (
                "regex_word_replace" in substitution
                and substitution["regex_word_replace"]
            ):
                df_selected = __apply_regex_substitution(
                    df_selected, column_name, substitution, spark.sparkContext
                )
            elif "substitution_file" in substitution:
                df_selected = __apply_substitution(
                    df_selected, column_name, substitution, spark.sparkContext
                )
            else:
                raise KeyError(
                    "You must supply a substitution file and either specify regex_word_replace=true or supply a join value."
                )
    return df_selected


def __load_substitutions(file_name):
    """Reads in the substitution file and returns a 2-tuple representing it.

    Parameters
    ----------
    file_name: name of substitution file

    Returns
    ----------
    A 2-tuple where the first value is an array of values to be replaced and the second is an array of values to use
    when replacing words in the first array.
    """
    sub_froms = []
    sub_tos = []
    with open(file_name, mode="r", encoding="utf-8-sig") as f:
        for line in f:
            sub_to, sub_from = line.strip().lower().split(",")
            sub_froms.append(sub_from)
            sub_tos.append(sub_to)
    return (sub_froms, sub_tos)


def __apply_substitution(df, column_name, substitution, sc):
    """Returns a new df with the values in the column column_name replaced using substitutions defined in substitution_file."""
    substitution_file = substitution["substitution_file"]
    join_value = substitution["join_value"]
    join_column = substitution["join_column"]
    join_column_alias = join_column + "_sub"
    sub_froms, sub_tos = __load_substitutions(substitution_file)
    subs = list(zip(sub_froms, sub_tos))
    Sub = namedtuple("Sub", ["sub_from", "sub_to"])
    sub_df = (
        sc.parallelize(subs, 1)
        .map(lambda s: Sub(s[0], s[1]))
        .toDF()
        .withColumn(join_column_alias, lit(join_value))
    )
    join_statement = (sub_df["sub_from"] == split(df[column_name], " ")[0]) & (
        sub_df[join_column_alias] == df[join_column]
    )
    df_sub = df.join(sub_df.hint("broadcast"), join_statement, "left_outer").drop(
        "join_column_alias"
    )
    df_sub_select = (
        when(df_sub["sub_to"].isNull(), df_sub[column_name])
        .otherwise(concat_ws(" ", df_sub["sub_to"], split(df_sub[column_name], " ")[1]))
        .alias(column_name)
    )
    df_sub_selects = list(set(df.columns) - set([column_name])) + [df_sub_select]
    return df_sub.select(df_sub_selects)


def __apply_regex_substitution(df, column_name, substitution, sc):
    """Returns a new df with the values in the column column_name replaced using substitutions defined in substitution_file."""

    substitution_file = substitution["substitution_file"]
    sub_froms, sub_tos = __load_substitutions(substitution_file)
    subs = dict(zip(sub_froms, sub_tos))
    col = column_name
    df.checkpoint()

    for sub_from, sub_to in subs.items():
        col = regexp_replace(
            col, r"(?:(?<=\s)|(?<=^))(" + sub_from + r")(?:(?=\s)|(?=$))", sub_to
        )
    return df.withColumn(column_name, col)



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from setuptools import setup, find_packages
import os


packages_with_templates = [
    "hlink.linking",
    "hlink.linking.preprocessing",
    "hlink.linking.matching",
    "hlink.linking.hh_matching",
    "hlink.linking.model_exploration",
    "hlink.linking.hh_model_exploration",
    "hlink.linking.reporting",
    "hlink.linking.training",
    "hlink.linking.hh_training",
]

package_data = {"hlink.spark": ["jars/hlink_lib-assembly-1.0.jar"]}
for package in packages_with_templates:
    package_path = package.replace(".", "/")
    template_files = []
    for root, dirs, files in os.walk(f"{package_path}/templates"):
        for file in files:
            template_files.append(
                os.path.relpath(os.path.join(root, file), package_path)
            )
    package_data[package] = template_files

package_data["hlink.linking"].append("table_definitions.csv")

install_requires = [
    "colorama==0.4.4",
    "ipython==8.3.0",
    "Jinja2==3.1.2",
    "numpy==1.22.3",
    "pandas==1.4.2",
    "pyspark==3.2.1",
    "scikit-learn==1.1.0",
    "toml==0.10.2",
]

dev_requires = [
    "pre-commit",
    "black==22.3.0",
    "flake8==4.0.1",
    "sphinx",
    "recommonmark",
    "pytest==7.1.2",
]

setup(
    name="hlink",
    version="3.2.6",
    packages=find_packages(),
    description="Fast supervised pyspark record linkage software",
    package_data=package_data,
    install_requires=install_requires,
    extras_require={"dev": dev_requires},
    project_urls={'Homepage': 'https://github.com/ipums/hlink'},
    entry_points="""
    [console_scripts]
    hlink=hlink.scripts.main:cli
  """,
)



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from hlink.spark.session import SparkConnection
import hlink.scripts.main_loop
import hlink.tests
from hlink.configs.load_config import load_conf_file
import hlink.linking.matching as match
import hlink.linking.training as train
import hlink.linking.hh_training as hh_train
import hlink.linking.hh_matching as hh_match
import hlink.linking.preprocessing as pre
import hlink.linking.reporting as rep
import hlink.linking.model_exploration as me
import hlink.linking.hh_model_exploration as hh_me
from hlink.linking.link_run import LinkRun
import json
import os
import pytest
import sys
from types import SimpleNamespace


pytest_plugins = (
    "hlink.tests.plugins.datasources",
    "hlink.tests.plugins.external_data_paths",
)


def load_table_from_csv(link_task, path, table_name):
    link_task.spark.read.csv(path, header=True, inferSchema=True).write.mode(
        "overwrite"
    ).saveAsTable(table_name)


@pytest.fixture(scope="session")
def spark(tmpdir_factory):
    os.environ["PYSPARK_PYTHON"] = sys.executable
    spark_connection = SparkConnection(
        tmpdir_factory.mktemp("derby"),
        tmpdir_factory.mktemp("warehouse"),
        tmpdir_factory.mktemp("spark_tmp_dir"),
        sys.executable,
        "linking",
    )
    return spark_connection.local()


@pytest.fixture(scope="function", autouse=True)
def set_shuffle(spark):
    spark.conf.set("spark.sql.shuffle.partitions", "1")
    spark.conf.set("spark.default.parallelism", "1")


@pytest.fixture(scope="function", autouse=True)
def drop_all_tables(main):
    main.do_drop_all("")


@pytest.fixture()
def package_path():
    """The path to the tests package."""
    return os.path.dirname(hlink.tests.conftest.__file__)


@pytest.fixture()
def input_data_dir_path(package_path):
    """The path to the directory containing test input data."""
    return os.path.join(package_path, "input_data")


@pytest.fixture()
def conf_dir_path(package_path):
    """The path to the directory containing test config files."""
    return os.path.join(package_path, "conf")


@pytest.fixture()
def spark_test_tmp_dir_path(spark, package_path):
    """The path to the test potential_matches csv file."""
    path = f"output_data/spark_test_tmp_dir{os.getenv('PYTEST_XDIST_WORKER', '')}"
    full_path = os.path.join(package_path, path)
    return full_path


@pytest.fixture()
def link_run(spark, conf):
    return LinkRun(spark, conf)


@pytest.fixture()
def preprocessing(link_run):
    return pre.Preprocessing(link_run)


@pytest.fixture()
def main(link_run):
    main = hlink.scripts.main_loop.Main(link_run)
    main.preloop()
    return main


@pytest.fixture()
def matching(link_run):
    return match.Matching(link_run)


@pytest.fixture()
def hh_matching(link_run):
    return hh_match.HHMatching(link_run)


@pytest.fixture()
def training(link_run):
    return train.Training(link_run)


@pytest.fixture()
def hh_training(link_run):
    return hh_train.HHTraining(link_run)


@pytest.fixture()
def reporting(link_run):
    return rep.Reporting(link_run)


@pytest.fixture()
def model_exploration(link_run):
    return me.ModelExploration(link_run)


@pytest.fixture()
def hh_model_exploration(link_run):
    return hh_me.HHModelExploration(link_run)


@pytest.fixture(scope="module")
def fake_self(spark):
    d = {"training_conf": "training"}
    n = SimpleNamespace(**d)
    return n


# Because of the way pytest fixtures work, `conf` is evaluated only once per test
# function call. The result is cached, and any subsequent requests for the fixture
# return the *same object*. Since this fixture returns a mutable dictionary, changes
# to the returned object will affect other fixtures that request and use `conf`.
#
# We use this with the `link_run` fixture and other fixtures like `preprocessing_conf`.
# Tests will request `preprocessing_conf`, which modifies `conf`. This modifies the
# `LinkRun.config` dictionary, since that is also a pointer to `conf`. Any additional
# modifications to `preprocessing_conf` in the test are also applied to `LinkRun.config`.
#
# TODO: Maybe think of a different way to do this. This way is convenient, but it can
# be hard to understand.
@pytest.fixture(scope="function")
def conf(conf_dir_path):
    return get_conf(conf_dir_path, "test.json")


@pytest.fixture(scope="function")
def integration_conf(input_data_dir_path, conf_dir_path):
    conf_file = os.path.join(conf_dir_path, "integration")
    conf = load_conf_file(conf_file)

    datasource_a = conf["datasource_a"]
    datasource_b = conf["datasource_b"]
    training = conf["training"]
    datasource_a["file"] = os.path.join(input_data_dir_path, datasource_a["file"])
    datasource_b["file"] = os.path.join(input_data_dir_path, datasource_b["file"])
    training["dataset"] = os.path.join(input_data_dir_path, training["dataset"])
    return conf


def get_conf(conf_dir_path, name):
    path_to_file = os.path.join(conf_dir_path, name)
    with open(path_to_file) as f:
        conf = json.load(f)
    return conf


@pytest.fixture(scope="function")
def preprocessing_conf(spark, conf, base_datasources):
    """Create a fixture to set the conf datasource_(a/b) values to the test data"""
    pathname_a, pathname_b = base_datasources
    conf["datasource_a"] = {"parquet_file": pathname_a}
    conf["datasource_b"] = {"parquet_file": pathname_b}
    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_all_space_columns(
    spark, conf, datasource_unrestricted_blank_columns
):
    """Create a fixture to set the conf datasource_(a/b) values to the test data"""
    pathname_a, pathname_b = datasource_unrestricted_blank_columns
    conf["datasource_a"] = {"parquet_file": pathname_a}
    conf["datasource_b"] = {"parquet_file": pathname_b}
    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_simple_names(spark, conf, datasource_preprocessing_simple_names):
    """Create a fixture for testing name substitution and bigrams"""
    pathname_a, pathname_b = datasource_preprocessing_simple_names
    conf["datasource_a"] = {"parquet_file": pathname_a}
    conf["datasource_b"] = {"parquet_file": pathname_b}
    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_popularity(spark, conf, ext_path_preprocessing_popularity):
    """Create a fixture for testing name substitution and bigrams"""
    pathname = ext_path_preprocessing_popularity
    conf["datasource_a"] = {"file": pathname}
    conf["datasource_b"] = {"file": pathname}

    conf["column_mappings"] = [
        {"column_name": "sex"},
        {"column_name": "namefrst"},
        {"column_name": "namelast"},
        {"column_name": "birthyr"},
        {"column_name": "bpl"},
    ]

    conf["blocking"] = []
    conf["comparisons"] = {}

    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_street_names(spark, conf, test_street_names_data_path):
    """Create a fixture for testing street name abbreviation substitutions"""
    conf["datasource_a"] = {"file": test_street_names_data_path}
    conf["datasource_b"] = {"file": test_street_names_data_path}
    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_birthyr(spark, conf, birthyr_replace_path):
    """Create a fixture for testing name substitution and bigrams"""
    conf["datasource_a"] = {"file": birthyr_replace_path}
    conf["datasource_b"] = {"file": birthyr_replace_path}
    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_synthetic_household_data(
    spark, conf, datasource_synthetic_households
):
    """Create a fixture conf for testing union transform of household/neighborhood data"""
    pathname_a, pathname_b = datasource_synthetic_households
    conf["datasource_a"] = {"parquet_file": pathname_a}
    conf["datasource_b"] = {"parquet_file": pathname_b}
    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_19thc_nativity_conf(
    spark, conf, datasource_19thc_nativity_households_data
):
    """Create a fixture conf for testing nativity calculation"""
    full_path_a, full_path_b = datasource_19thc_nativity_households_data
    conf["datasource_a"] = {"file": full_path_a}
    conf["datasource_b"] = {"file": full_path_b}

    conf["column_mappings"] = [
        {"column_name": "serial"},
        {"column_name": "pernum"},
        {"column_name": "relate"},
        {"column_name": "bpl"},
        {"column_name": "momloc"},
        {"column_name": "poploc"},
        {"column_name": "namefrst"},
        {"column_name": "namelast"},
        {"column_name": "key_nativity_calc"},
        {
            "column_name": "nativity",
            "set_value_column_b": 0,
        },
        {
            "alias": "test_nativity",
            "column_name": "key_nativity_calc",
            "set_value_column_a": 0,
        },
        {"column_name": "key_mbpl"},
        {"column_name": "key_fbpl"},
        {
            "column_name": "mbpl",
            "set_value_column_a": 999,
        },
        {
            "column_name": "key_mbpl_range",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_mbpl_range_b",
            "set_value_column_a": 0,
        },
        {
            "column_name": "key_mother_nativity",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_mbpl_match",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_fbpl_match",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_mfbpl_match",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_m_caution_1870_1880",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_m_caution_1850_1860",
            "set_value_column_b": 0,
        },
    ]

    conf["blocking"] = [{"column_name": "id"}]
    conf["comparisons"] = {}

    conf["feature_selections"] = [
        {
            "output_column": "mbpl_range_b",
            "input_col": "mbpl",
            "transform": "sql_condition",
            "condition": "case when mbpl >= 997 then 0 when mbpl < 100 then 1 when (mbpl > 99 and mbpl < 997) then 2 else 0 end",
            "set_value_column_a": 0,
        },
        {
            "family_id": "serial",
            "other_col": "nativity",
            "output_col": "mother_nativity",
            "person_id": "pernum",
            "person_pointer": "momloc",
            "transform": "attach_family_col",
            "set_value_column_b": 0,
        },
        {
            "output_col": "mbpl_calc",
            "transform": "attach_family_col",
            "other_col": "bpl",
            "person_pointer": "momloc",
            "family_id": "serial",
            "person_id": "pernum",
        },
        {
            "output_col": "fbpl_calc",
            "transform": "attach_family_col",
            "other_col": "bpl",
            "person_pointer": "poploc",
            "family_id": "serial",
            "person_id": "pernum",
        },
        {
            "post_agg_feature": True,
            "output_column": "mbpl_range",
            "input_column": "mother_nativity",
            "transform": "sql_condition",
            "condition": "case when mother_nativity == 0 then 0 when mother_nativity > 0 and mother_nativity < 5 then 1 when mother_nativity == 5 then 2 else 0 end",
            "override_column_b": "mbpl_range_b",
        },
        {
            "post_agg_feature": True,
            "output_column": "nativity_calc",
            "transform": "sql_condition",
            "input_columns": ["bpl", "mbpl_calc", "fbpl_calc"],
            "condition": """case
                            when bpl >= 997 or mbpl_calc >= 997 or fbpl_calc >= 997
                            then 0
                            when bpl < 100 and fbpl_calc < 100 and mbpl_calc < 100
                            then 1
                            when bpl < 100 and mbpl_calc > 99 and fbpl_calc > 99
                            then 4
                            when bpl < 100 and mbpl_calc > 99 and fbpl_calc < 100
                            then 3
                            when bpl < 100 and fbpl_calc > 99 and mbpl_calc < 100
                            then 2
                            when bpl > 99 and bpl < 997
                            then 5
                            else 0
                            end""",
            "override_column_b": "test_nativity",
        },
    ]

    conf["comparison_features"] = [
        {
            "alias": "mbpl_match",
            "column_name": "mbpl_calc",
            "comparison_type": "present_and_matching_categorical",
        },
        {
            "alias": "fbpl_match",
            "column_name": "fbpl_calc",
            "comparison_type": "present_and_matching_categorical",
        },
        {
            "alias": "mfbpl_match",
            "column_name": "nativity_calc",
            "comparison_type": "present_and_equal_categorical_in_universe",
            "NIU": "0",
            "categorical": True,
        },
        {
            "alias": "m_caution_1870_1880",
            "categorical": True,
            "column_name": "mbpl_range",
            "comparison_type": "not_zero_and_not_equals",
        },
        {
            "alias": "m_caution_1850_1860",
            "categorical": True,
            "column_name": "mbpl_calc",
            "comparison_type": "present_and_not_equal",
        },
        {
            "column_name": "key_mbpl_match",
            "alias": "key_mbpl_match",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_fbpl_match",
            "alias": "key_fbpl_match",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_mfbpl_match",
            "alias": "key_mfbpl_match",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_m_caution_1870_1880",
            "alias": "key_m_caution_1870_1880",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_m_caution_1850_1860",
            "alias": "key_m_caution_1850_1860",
            "comparison_type": "fetch_a",
        },
    ]

    conf["training"] = {
        "dependent_var": "match",
        "independent_vars": [
            "mbpl_match",
            "fbpl_match",
            "mfbpl_match",
            "m_caution_1870_1880",
            "m_caution_1850_1860",
        ],
    }

    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_19thc_caution_conf(
    spark, conf, datasource_19thc_nativity_households_data
):
    """Create a fixture conf for testing nativity calculation"""
    full_path_a, full_path_b = datasource_19thc_nativity_households_data
    conf["datasource_a"] = {"file": full_path_a}
    conf["datasource_b"] = {"file": full_path_b}

    conf["column_mappings"] = [
        {"column_name": "serial"},
        {"column_name": "pernum"},
        {"column_name": "relate"},
        {"column_name": "bpl"},
        {"column_name": "momloc"},
        {"column_name": "stepmom"},
        {"column_name": "poploc"},
        {"column_name": "namefrst"},
        {"column_name": "namelast"},
        {"column_name": "birthyr"},
        {"column_name": "key_nativity_calc"},
        {
            "column_name": "nativity",
            "set_value_column_b": 0,
        },
        {
            "alias": "test_nativity",
            "column_name": "key_nativity_calc",
            "set_value_column_a": 0,
        },
        {"column_name": "key_mbpl"},
        {"column_name": "key_fbpl"},
        {
            "column_name": "mbpl",
            "set_value_column_a": 999,
        },
        {
            "column_name": "key_mbpl_range",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_mbpl_range_b",
            "set_value_column_a": 0,
        },
        {
            "column_name": "key_mother_nativity",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_mbpl_match",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_fbpl_match",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_mfbpl_match",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_m_caution_1870_1880",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_m_caution_1850_1860",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_m_caution_cc3_012",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_m_caution_cc4_012",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_intermediate_mbpl_range_not_equals",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_intermediate_mbpl_range_not_zero_and_not_equals",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_intermediate_mother_birthyr_abs_diff_5",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_intermediate_stepmom_parent_step_change",
            "set_value_column_b": 0,
        },
        {
            "column_name": "key_intermediate_momloc_present_both_years",
            "set_value_column_b": 0,
        },
    ]

    conf["blocking"] = [{"column_name": "id"}]
    conf["comparisons"] = {}

    conf["feature_selections"] = [
        {
            "output_column": "mbpl_range_b",
            "input_col": "mbpl",
            "transform": "sql_condition",
            "condition": "case when mbpl >= 997 then 0 when mbpl < 100 then 1 when (mbpl > 99 and mbpl < 997) then 2 else 0 end",
            "set_value_column_a": 0,
        },
        {
            "family_id": "serial",
            "other_col": "nativity",
            "output_col": "mother_nativity",
            "person_id": "pernum",
            "person_pointer": "momloc",
            "transform": "attach_family_col",
            "set_value_column_b": 0,
        },
        {
            "output_col": "mbpl_calc",
            "transform": "attach_family_col",
            "other_col": "bpl",
            "person_pointer": "momloc",
            "family_id": "serial",
            "person_id": "pernum",
        },
        {
            "output_col": "fbpl_calc",
            "transform": "attach_family_col",
            "other_col": "bpl",
            "person_pointer": "poploc",
            "family_id": "serial",
            "person_id": "pernum",
        },
        {
            "output_col": "mother_birthyr",
            "transform": "attach_family_col",
            "other_col": "birthyr",
            "person_pointer": "momloc",
            "family_id": "serial",
            "person_id": "pernum",
        },
        {
            "post_agg_feature": True,
            "output_column": "mbpl_range",
            "input_column": "mother_nativity",
            "transform": "sql_condition",
            "condition": "case when mother_nativity == 0 then 0 when mother_nativity > 0 and mother_nativity < 5 then 1 when mother_nativity == 5 then 2 else 0 end",
            "override_column_b": "mbpl_range_b",
        },
        {
            "post_agg_feature": True,
            "output_column": "nativity_calc",
            "transform": "sql_condition",
            "input_columns": ["bpl", "mbpl_calc", "fbpl_calc"],
            "condition": """case
                            when bpl >= 997 or mbpl_calc >= 997 or fbpl_calc >= 997
                            then 0
                            when bpl < 100 and fbpl_calc < 100 and mbpl_calc < 100
                            then 1
                            when bpl < 100 and mbpl_calc > 99 and fbpl_calc > 99
                            then 4
                            when bpl < 100 and mbpl_calc > 99 and fbpl_calc < 100
                            then 3
                            when bpl < 100 and fbpl_calc > 99 and mbpl_calc < 100
                            then 2
                            when bpl > 99 and bpl < 997
                            then 5
                            else 0
                            end""",
            "override_column_b": "test_nativity",
        },
    ]

    conf["comparison_features"] = [
        {
            "alias": "mbpl_match",
            "column_name": "mbpl_calc",
            "comparison_type": "present_and_matching_categorical",
        },
        {
            "alias": "fbpl_match",
            "column_name": "fbpl_calc",
            "comparison_type": "present_and_matching_categorical",
        },
        {
            "alias": "mfbpl_match",
            "column_name": "nativity_calc",
            "comparison_type": "present_and_equal_categorical_in_universe",
            "NIU": "0",
            "categorical": True,
        },
        {
            "alias": "m_caution_1870_1880",
            "categorical": True,
            "column_name": "mbpl_range",
            "comparison_type": "not_zero_and_not_equals",
        },
        {
            "alias": "m_caution_1850_1860",
            "categorical": True,
            "column_name": "mbpl_calc",
            "comparison_type": "present_and_not_equal",
        },
        {
            "column_name": "key_mbpl_match",
            "alias": "key_mbpl_match",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_fbpl_match",
            "alias": "key_fbpl_match",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_mfbpl_match",
            "alias": "key_mfbpl_match",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_m_caution_1870_1880",
            "alias": "key_m_caution_1870_1880",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_m_caution_1850_1860",
            "alias": "key_m_caution_1850_1860",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_m_caution_cc3_012",
            "alias": "key_m_caution_cc3_012",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_m_caution_cc4_012",
            "alias": "key_m_caution_cc4_012",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_intermediate_mbpl_range_not_equals",
            "alias": "key_intermediate_mbpl_range_not_equals",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_intermediate_mbpl_range_not_zero_and_not_equals",
            "alias": "key_intermediate_mbpl_range_not_zero_and_not_equals",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_intermediate_mother_birthyr_abs_diff_5",
            "alias": "key_intermediate_mother_birthyr_abs_diff_5",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_intermediate_stepmom_parent_step_change",
            "alias": "key_intermediate_stepmom_parent_step_change",
            "comparison_type": "fetch_a",
        },
        {
            "column_name": "key_intermediate_momloc_present_both_years",
            "alias": "key_intermediate_momloc_present_both_years",
            "comparison_type": "fetch_a",
        },
        {
            "alias": "m_caution_cc3_012",
            "column_names": ["mbpl_range", "mother_birthyr", "momloc"],
            "comparison_type": "caution_comp_3_012",
            "categorical": True,
            "comp_a": {
                "column_name": "mbpl_range",
                "comparison_type": "not_equals",
            },
            "comp_b": {
                "column_name": "mother_birthyr",
                "comparison_type": "abs_diff",
                "gt_threshold": 5,
            },
            "comp_c": {
                "column_name": "momloc",
                "comparison_type": "present_both_years",
            },
        },
        {
            "alias": "m_caution_cc4_012",
            "column_names": ["mbpl_range", "mother_birthyr", "stepmom", "momloc"],
            "comparison_type": "caution_comp_4_012",
            "categorical": True,
            "comp_a": {
                "column_name": "mbpl_range",
                "comparison_type": "not_zero_and_not_equals",
            },
            "comp_b": {
                "column_name": "mother_birthyr",
                "comparison_type": "abs_diff",
                "gt_threshold": 5,
            },
            "comp_c": {
                "column_name": "stepmom",
                "comparison_type": "parent_step_change",
            },
            "comp_d": {
                "column_name": "momloc",
                "comparison_type": "present_both_years",
            },
        },
        {
            "alias": "intermediate_mbpl_range_not_equals",
            "column_name": "mbpl_range",
            "comparison_type": "not_equals",
        },
        {
            "alias": "intermediate_mbpl_range_not_zero_and_not_equals",
            "column_name": "mbpl_range",
            "comparison_type": "not_zero_and_not_equals",
        },
        {
            "alias": "intermediate_mother_birthyr_abs_diff_5",
            "column_name": "mother_birthyr",
            "comparison_type": "abs_diff",
            "gt_threshold": 5,
        },
        {
            "alias": "intermediate_stepmom_parent_step_change",
            "column_name": "stepmom",
            "comparison_type": "parent_step_change",
        },
        {
            "alias": "intermediate_momloc_present_both_years",
            "column_name": "momloc",
            "comparison_type": "present_both_years",
        },
    ]

    conf["training"] = {
        "dependent_var": "match",
        "independent_vars": [
            "mbpl_match",
            "fbpl_match",
            "mfbpl_match",
            "m_caution_1870_1880",
            "m_caution_1850_1860",
            "m_caution_cc3_012",
            "m_caution_cc4_012",
            "intermediate_mbpl_range_not_equals",
            "intermediate_mbpl_range_not_zero_and_not_equals",
            "intermediate_mother_birthyr_abs_diff_5",
            "intermediate_stepmom_parent_step_change",
            "intermediate_momloc_present_both_years",
        ],
    }

    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_household_data(spark, conf, datasource_real_households):
    """Create a fixture conf for testing family/neighborhood transforms"""
    full_path_a, full_path_b = datasource_real_households
    conf["datasource_a"] = {"file": full_path_a}
    conf["datasource_b"] = {"file": full_path_b}

    conf["column_mappings"] = [
        {"column_name": "namefrst", "alias": "namefrst_orig"},
        {"column_name": "namelast", "alias": "namelast_orig"},
        {"column_name": "bpl"},
        {"column_name": "sex"},
        {"column_name": "enumdist"},
        {"column_name": "pernum"},
        {"column_name": "serial"},
        {"column_name": "sploc"},
        {"column_name": "poploc"},
        {"column_name": "momloc"},
        {"column_name": "relate"},
        {
            "column_name": "namefrst",
            "alias": "namefrst_clean",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {
                    "type": "remove_suffixes",
                    "values": ["jr", "sr", "ii", "iii", "iv", "v", "vi", "vii", "viii"],
                },
                {"type": "remove_alternate_names"},
                {"type": "condense_strip_whitespace"},
            ],
        },
        {
            "column_name": "namelast",
            "alias": "namelast_clean",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {
                    "type": "remove_suffixes",
                    "values": ["jr", "sr", "ii", "iii", "iv", "v", "vi", "vii", "viii"],
                },
                {"type": "remove_prefixes", "values": ["ah"]},
                {
                    "type": "condense_prefixes",
                    "values": ["mc", "mac", "o", "de", "van", "di"],
                },
                {"type": "remove_one_letter_names"},
                {"type": "remove_alternate_names"},
                {"type": "condense_strip_whitespace"},
            ],
        },
    ]

    conf["blocking"] = []
    conf["comparisons"] = {}

    return conf


@pytest.fixture(scope="function")
def preprocessing_conf_rel_rows(spark, conf, test_data_rel_rows_age):
    """Create a fixture conf for testing family/neighborhood transforms"""
    full_path_a, full_path_b = test_data_rel_rows_age
    conf["datasource_a"] = {"file": full_path_a}
    conf["datasource_b"] = {"file": full_path_b}
    conf["id_column"] = "histid"

    conf["column_mappings"] = [
        {"column_name": "namefrst", "alias": "namefrst_orig"},
        {"column_name": "namelast", "alias": "namelast_orig"},
        {"column_name": "birthyr"},
        {"column_name": "sex"},
        {"column_name": "pernum"},
        {"column_name": "serialp"},
        {"column_name": "relate"},
        {"column_name": "age"},
        {"column_name": "yearp", "alias": "year"},
        {
            "column_name": "namefrst",
            "alias": "namefrst_clean",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {"type": "remove_suffixes", "values": ["jr", "sr", "ii", "iii"]},
                {"type": "remove_alternate_names"},
                {"type": "condense_strip_whitespace"},
            ],
        },
        {
            "column_name": "namefrst_clean",
            "alias": "namefrst_split",
            "transforms": [{"type": "split"}],
        },
        {
            "column_name": "namefrst_split",
            "alias": "namefrst_unstd",
            "transforms": [{"type": "array_index", "value": 0}],
        },
        {
            "column_name": "birthyr",
            "alias": "clean_birthyr",
            "transforms": [
                {
                    "type": "mapping",
                    "mappings": {9999: "", 1999: ""},
                    "output_type": "int",
                }
            ],
        },
    ]

    conf["feature_selections"] = [
        {
            "input_column": "clean_birthyr",
            "output_column": "replaced_birthyr",
            "condition": "case when clean_birthyr is null or clean_birthyr == '' then year - age else clean_birthyr end",
            "transform": "sql_condition",
        },
        {
            "family_id": "serialp",
            "input_cols": [
                "histid",
                "namefrst_unstd",
                "replaced_birthyr",
                "sex",
                "relate",
            ],
            "output_col": "namefrst_related_rows",
            "transform": "related_individual_rows",
            "filters": [
                {"column": "relate", "min": 300, "max": 1099},
                {"column": "age", "min": 0, "max": 999},
            ],
        },
        {
            "family_id": "serialp",
            "input_cols": [
                "histid",
                "namefrst_unstd",
                "replaced_birthyr",
                "sex",
                "relate",
            ],
            "output_col": "namefrst_related_rows_age_min_5",
            "transform": "related_individual_rows",
            "filters": [
                {"column": "relate", "min": 300, "max": 1099},
                {"column": "age", "min": 5, "max": 999},
            ],
        },
        {
            "family_id": "serialp",
            "input_cols": [
                "histid",
                "namefrst_unstd",
                "replaced_birthyr",
                "sex",
                "relate",
            ],
            "output_col": "namefrst_related_rows_age_b_min_5",
            "transform": "related_individual_rows",
            "filters": [
                {"column": "relate", "min": 300, "max": 1099},
                {"column": "age", "min": 5, "max": 999, "dataset": "b"},
            ],
        },
    ]

    conf["blocking"] = []
    conf["comparisons"] = {}

    return conf


@pytest.fixture(scope="function")
def matching_conf(spark, conf, datasource_matching, matching):
    """Create conf fixture for testing matching steps using the prepped_df_(a/b) dataframes and populate basic config values"""
    matching.link_run.print_sql = True
    conf["column_mappings"] = [
        {"column_name": "serialp"},
        {"column_name": "namefrst"},
        {"column_name": "namelast"},
        {"column_name": "bpl"},
        {"column_name": "sex"},
        {"column_name": "street"},
        {"column_name": "enum_dist"},
    ]
    conf["blocking"] = [{"column_name": "sex"}]
    conf["comparisons"] = {
        "comp_a": {
            "feature_name": "namelast_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "comp_b": {
            "feature_name": "namelast_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "operator": "AND",
    }
    conf["training"] = {}

    conf["training"]["dependent_var"] = "match"
    conf["training"]["independent_vars"] = ["namefrst_jw", "namelast_jw", "ssex"]
    return conf


@pytest.fixture(scope="function")
def matching_conf_namefrst_std_and_unstd(
    spark, conf, matching, test_data_blocking_double_comparison
):
    """Create conf fixture for testing matching steps using the prepped_df_(a/b) dataframes and populate basic config values"""

    conf["id_column"] = "histid"
    full_path_a, full_path_b = test_data_blocking_double_comparison
    conf["datasource_a"] = {"file": full_path_a}
    conf["datasource_b"] = {"file": full_path_b}

    conf["column_mappings"] = [
        {"column_name": "namefrst_unstd"},
        {"column_name": "namefrst_std"},
        {"column_name": "namelast_clean"},
        {"column_name": "bpl_clean"},
        {"column_name": "sex"},
        {"column_name": "birthyr"},
    ]
    conf["blocking"] = [
        {"column_name": "sex"},
        {"column_name": "bpl_clean"},
        {"column_name": "birthyr"},
    ]

    conf["comparisons"] = {
        "operator": "AND",
        "comp_a": {
            "operator": "OR",
            "comp_a": {
                "feature_name": "namefrst_unstd_jw",
                "threshold": 0.70,
                "comparison_type": "threshold",
            },
            "comp_b": {
                "feature_name": "namefrst_std_jw",
                "threshold": 0.70,
                "comparison_type": "threshold",
            },
        },
        "comp_b": {
            "feature_name": "namelast_jw",
            "threshold": 0.70,
            "comparison_type": "threshold",
        },
    }

    conf["comparison_features"] = [
        {
            "alias": "namefrst_unstd_jw",
            "column_name": "namefrst_unstd",
            "comparison_type": "jaro_winkler",
        },
        {
            "alias": "namefrst_std_jw",
            "column_name": "namefrst_std",
            "comparison_type": "jaro_winkler",
        },
        {
            "alias": "namelast_jw",
            "column_name": "namelast_clean",
            "comparison_type": "jaro_winkler",
        },
    ]

    conf["training"] = {}
    return conf


@pytest.fixture(scope="function")
def blocking_explode_conf(spark, conf):
    """Create conf fixture for testing matching steps using the family/neighborhood transforms and populate basic config values"""

    conf["column_mappings"] = [
        {"column_name": "namefrst"},
        {"column_name": "namelast"},
        {"column_name": "birthyr"},
        {"column_name": "sex"},
    ]

    conf["blocking"] = [
        {
            "column_name": "birthyr_3",
            "dataset": "a",
            "derived_from": "birthyr",
            "expand_length": 3,
            "explode": True,
        },
        {"column_name": "sex"},
    ]

    conf["comparison_features"] = [
        {
            "alias": "namefrst_jw",
            "column_name": "namefrst",
            "comparison_type": "jaro_winkler",
        },
        {
            "alias": "namelast_jw",
            "column_name": "namelast",
            "comparison_type": "jaro_winkler",
        },
        {
            "alias": "ssex",
            "column_name": "sex",
            "comparison_type": "equals",
            "categorical": True,
        },
    ]
    conf["comparisons"] = {
        "comp_a": {
            "feature_name": "namefrst_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "comp_b": {
            "feature_name": "namelast_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "operator": "AND",
    }

    conf["training"] = {}
    conf["training"]["dependent_var"] = "match"
    conf["training"]["independent_vars"] = ["namefrst_jw", "namelast_jw", "ssex"]
    return conf


@pytest.fixture(scope="function")
def matching_household_conf(
    spark, conf, datasource_real_households, preprocessing, matching
):
    """Create conf fixture for testing matching steps using the family/neighborhood transforms and populate basic config values"""

    full_path_a, full_path_b = datasource_real_households
    conf["datasource_a"] = {"file": full_path_a}
    conf["datasource_b"] = {"file": full_path_b}

    conf["column_mappings"] = [
        {"column_name": "namefrst", "alias": "namefrst_orig"},
        {"column_name": "namelast", "alias": "namelast_orig"},
        {"column_name": "bpl"},
        {"column_name": "birthyr"},
        {"column_name": "age"},
        {"column_name": "sex"},
        {"column_name": "enumdist"},
        {"column_name": "pernum"},
        {"column_name": "serial", "alias": "serialp"},
        {"column_name": "sploc"},
        {"column_name": "poploc"},
        {"column_name": "momloc"},
        {"column_name": "relate"},
        {
            "column_name": "namefrst",
            "alias": "namefrst_std",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {
                    "type": "remove_suffixes",
                    "values": ["jr", "sr", "ii", "iii", "iv", "v", "vi", "vii", "viii"],
                },
                {"type": "remove_alternate_names"},
                {"type": "condense_strip_whitespace"},
            ],
        },
        {
            "column_name": "namelast",
            "alias": "namelast_clean",
            "transforms": [
                {"type": "lowercase_strip"},
                {"type": "rationalize_name_words"},
                {"type": "remove_qmark_hyphen"},
                {"type": "replace_apostrophe"},
                {
                    "type": "remove_suffixes",
                    "values": ["jr", "sr", "ii", "iii", "iv", "v", "vi", "vii", "viii"],
                },
                {"type": "remove_prefixes", "values": ["ah"]},
                {
                    "type": "condense_prefixes",
                    "values": ["mc", "mac", "o", "de", "van", "di"],
                },
                {"type": "remove_one_letter_names"},
                {"type": "remove_alternate_names"},
                {"type": "condense_strip_whitespace"},
            ],
        },
    ]

    conf["blocking"] = [{"column_name": "sex"}]

    conf["comparisons"] = {
        "comp_a": {
            "feature_name": "namelast_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "comp_b": {
            "feature_name": "namelast_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "operator": "AND",
    }

    conf["training"] = {}
    conf["training"]["dependent_var"] = "match"

    return conf


@pytest.fixture(scope="function")
def matching_comparison_conf(spark, conf, datasource_matching_comparisons, matching):
    """Create conf fixture for testing matching steps using the prepped_df_(a/b) dataframes and populate basic config values"""
    conf["column_mappings"] = [
        {"column_name": "id"},
        {"column_name": "namelast"},
        {"column_name": "sex"},
        {"column_name": "mpbl"},
        {"column_name": "mother_birthyr"},
        {"column_name": "stepmom"},
        {"column_name": "spouse_bpl"},
        {"column_name": "spouse_birthyr"},
        {"column_name": "durmarr"},
        {"column_name": "mother_namefrst"},
        {"column_name": "spouse_namefrst"},
        {"column_name": "momloc"},
        {"column_name": "sploc"},
    ]
    conf["blocking"] = [{"column_name": "sex"}]
    conf["comparisons"] = {
        "comp_a": {
            "feature_name": "namelast_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "comp_b": {
            "feature_name": "namelast_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "operator": "AND",
    }

    return conf


@pytest.fixture(scope="function")
def matching_conf_counties(spark, conf, county_dist_datasources):
    """Create a fixture for testing name substitution and bigrams"""
    pathname_a, pathname_b = county_dist_datasources
    conf["datasource_a"] = {"parquet_file": pathname_a}
    conf["datasource_b"] = {"parquet_file": pathname_b}

    return conf


@pytest.fixture(scope="function")
def matching_conf_nativity(spark, conf, nativity_datasources):
    """Create a fixture for testing name substitution and bigrams"""
    pathname_a, pathname_b = nativity_datasources
    conf["datasource_a"] = {"file": pathname_a}
    conf["datasource_b"] = {"file": pathname_b}
    conf["training"] = {}

    conf["training"]["dependent_var"] = "match"

    return conf


@pytest.fixture(scope="function")
def training_conf(spark, conf, training_data_path, datasource_training):
    """Create the prepped_df_(a/b) dataframes and populate basic config values"""
    conf["training"] = {
        "dataset": training_data_path,
        "dependent_var": "match",
        "n_training_iterations": 10,
    }
    conf["column_mappings"] = [
        {"column_name": "serialp"},
        {"column_name": "namelast"},
        {"column_name": "bpl"},
        {"column_name": "sex"},
        {"column_name": "region"},
    ]
    conf["blocking"] = [{"column_name": "sex"}]
    conf["comparisons"] = {
        "comp_a": {
            "feature_name": "namelast_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "comp_b": {
            "feature_name": "namelast_jw",
            "threshold": 0.8,
            "comparison_type": "threshold",
        },
        "operator": "AND",
    }

    return conf


@pytest.fixture(scope="function")
def hh_training_conf(spark, conf, hh_training_data_path):
    """Create the prepped_df_(a/b) dataframes and populate basic config values"""
    conf["id_column"] = "histid"
    conf["drop_data_from_scored_matches"] = False
    conf["hh_training"] = {
        "dataset": hh_training_data_path,
        "dependent_var": "match",
        "prediction_col": "match",
        "n_training_iterations": 4,
        "seed": 120,
        "independent_vars": [
            "namelast_jw",
            "namefrst_jw",
            "byrdiff",
            "ssex",
            "srelate",
        ],
        "score_with_model": True,
        "use_training_data_features": False,
        "decision": "drop_duplicate_with_threshold_ratio",
        "get_precision_recall_curve": True,
        "chosen_model": {
            "type": "logistic_regression",
            "threshold": 0.5,
            "threshold_ratio": 1.2,
        },
        "model_parameters": [
            {"type": "logistic_regression", "threshold": 0.5, "threshold_ratio": 1.2},
            {
                "type": "random_forest",
                "maxDepth": 5.0,
                "numTrees": 75.0,
                "threshold": 0.5,
                "threshold_ratio": 1.2,
            },
        ],
    }
    conf["column_mappings"] = [
        {"column_name": "serialp"},
        {"column_name": "relate"},
        {"column_name": "namelast_clean"},
        {"column_name": "namefrst_unstd"},
        {"column_name": "clean_birthyr"},
        {"column_name": "sex"},
    ]
    conf["blocking"] = []
    conf["comparison_features"] = [
        {
            "alias": "byrdiff",
            "column_name": "clean_birthyr",
            "comparison_type": "abs_diff",
        },
        {
            "alias": "ssex",
            "column_name": "sex",
            "comparison_type": "equals",
            "categorical": True,
        },
        {
            "alias": "srelate",
            "column_name": "relate",
            "comparison_type": "equals",
            "categorical": True,
        },
        {
            "alias": "namefrst_jw",
            "column_name": "namefrst_unstd",
            "comparison_type": "jaro_winkler",
        },
        {
            "alias": "namelast_jw",
            "column_name": "namelast_clean",
            "comparison_type": "jaro_winkler",
        },
    ]

    return conf


@pytest.fixture(scope="function")
def hh_agg_feat_conf(spark, conf, hh_training_data_path):
    """Create the prepped_df_(a/b) dataframes and populate basic config values"""
    conf["id_column"] = "histid"
    conf["drop_data_from_scored_matches"] = False
    conf["hh_training"] = {
        "dataset": hh_training_data_path,
        "dependent_var": "match",
        "independent_vars": [
            "jw_max_a",
            "jw_max_b",
            "f1_match",
            "f2_match",
            "byrdiff",
            "sexmatch",
            "mardurmatch",
        ],
        "score_with_model": True,
        "chosen_model": {
            "type": "logistic_regression",
            "threshold": 0.5,
            "threshold_ratio": 1.2,
        },
    }
    conf["column_mappings"] = [
        {"column_name": "serialp"},
        {"column_name": "relate"},
        {"column_name": "pernum"},
        {"column_name": "namelast_clean"},
        {"column_name": "namefrst_unstd"},
        {"column_name": "clean_birthyr"},
        {"column_name": "sex"},
        {"column_name": "namefrst_mid_init"},
        {"column_name": "namefrst_init"},
        {"column_name": "namefrst_mid_init_2"},
    ]
    conf["blocking"] = []
    conf["comparison_features"] = [
        {
            "alias": "byrdiff",
            "column_name": "clean_birthyr",
            "comparison_type": "abs_diff",
        },
        {
            "alias": "sexmatch",
            "column_name": "sex",
            "comparison_type": "equals",
            "categorical": True,
        },
        {
            "alias": "namefrst_jw",
            "column_name": "namefrst_unstd",
            "comparison_type": "jaro_winkler",
        },
        {
            "alias": "namelast_jw",
            "column_name": "namelast_clean",
            "comparison_type": "jaro_winkler",
        },
        {"alias": "durmarr_a", "column_name": "durmarr", "comparison_type": "fetch_a"},
        {"alias": "durmarr_b", "column_name": "durmarr", "comparison_type": "fetch_b"},
        {
            "alias": "mardurmatch",
            "column_name": "durmarr",
            "not_equals": 99,
            "comparison_type": "abs_diff",
            "btwn_threshold": [9, 14],
            "categorical": True,
        },
        {
            "alias": "f1_match",
            "first_init_col": "namefrst_init",
            "mid_init_cols": ["namefrst_mid_init", "namefrst_mid_init_2"],
            "comparison_type": "f1_match",
            "categorical": True,
        },
        {
            "alias": "f2_match",
            "first_init_col": "namefrst_init",
            "mid_init_cols": ["namefrst_mid_init", "namefrst_mid_init_2"],
            "comparison_type": "f2_match",
            "categorical": True,
        },
        {
            "alias": "fn_a",
            "column_name": "namefrst_unstd",
            "comparison_type": "fetch_a",
        },
        {"alias": "fi_a", "column_name": "namefrst_init", "comparison_type": "fetch_a"},
        {
            "alias": "fn_b",
            "column_name": "namefrst_unstd",
            "comparison_type": "fetch_b",
        },
        {"alias": "fi_b", "column_name": "namefrst_init", "comparison_type": "fetch_b"},
        {
            "alias": "mi_a",
            "column_name": "namefrst_mid_init",
            "comparison_type": "fetch_a",
        },
        {
            "alias": "mi_b",
            "column_name": "namefrst_mid_init",
            "comparison_type": "fetch_b",
        },
    ]

    return conf



from math import ceil


def spark_shuffle_partitions_heuristic(dataset_size):
    """Calculate how many partitions to request from Spark based on dataset size.

    This is a heuristic / approximation of how many partitions should be requested
    from Spark so that hlink performs well. The minimum number of partitions
    returned is 200, the default for Spark.
    """
    partitions_approx = ceil(dataset_size / 25000)
    return max(200, partitions_approx)



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import logging
import pyspark.sql.functions as f
from pyspark.sql.window import Window

from hlink.linking.util import spark_shuffle_partitions_heuristic

from hlink.linking.link_step import LinkStep


class LinkStepReportR2PercentLinked(LinkStep):
    def __init__(self, task):
        super().__init__(
            task,
            "report round 2 percent linked",
            input_table_names=[
                "prepped_df_a",
                "predicted_matches",
                "hh_predicted_matches",
            ],
            output_table_names=[],
        )

    def _run(self):
        """For households with anyone linked in round 1, report percent of remaining household members linked in round 2."""

        dataset_size1 = self.task.spark.table("prepped_df_a").count()
        dataset_size2 = self.task.spark.table("predicted_matches").count()
        dataset_size3 = self.task.spark.table("hh_predicted_matches").count()
        dataset_size_max = max(dataset_size1, dataset_size2, dataset_size3)
        num_partitions = spark_shuffle_partitions_heuristic(dataset_size_max)
        self.task.spark.sql(f"set spark.sql.shuffle.partitions={num_partitions}")
        logging.info(
            f"Dataset sizes are {dataset_size1}, {dataset_size2}, {dataset_size3}, so set Spark partitions to {num_partitions} for this step"
        )

        pdfa = self.task.spark.table("prepped_df_a").select("serialp", "histid")
        pm = (
            self.task.spark.table("predicted_matches")
            .select("histid_a")
            .withColumn("linked_round", f.lit(1))
        )
        hhpm = (
            self.task.spark.table("hh_predicted_matches")
            .select("histid_a")
            .withColumn("linked_round", f.lit(2))
        )
        linked_rnds = (
            pdfa.join(pm, pdfa["histid"] == pm["histid_a"], "left")
            .drop("histid_a")
            .join(hhpm, pdfa["histid"] == hhpm["histid_a"], "left")
            .drop("histid_a")
            .select(
                "serialp",
                "histid",
                f.when(~f.isnull(pm["linked_round"]), pm["linked_round"])
                .otherwise(hhpm["linked_round"])
                .alias("linked_round"),
            )
            .fillna(0)
        )

        linked_rnds.cache().createOrReplaceTempView("linked_rounds")

        window = Window.partitionBy(linked_rnds["serialp"])
        df = linked_rnds.withColumn("histid_ct_total", f.count("serialp").over(window))
        df0 = df.withColumn(
            "R1count", f.count(f.when(f.col("linked_round") == 1, True)).over(window)
        )
        df1 = df0.withColumn(
            "R2count", f.count(f.when(f.col("linked_round") == 2, True)).over(window)
        )

        dfu = df1.select("serialp", "histid_ct_total", "R1count", "R2count").distinct()
        df2 = dfu.withColumn("R1_pct", dfu["R1count"] / dfu["histid_ct_total"])
        df3 = df2.withColumn(
            "R2_pct", df2["R2count"] / (df2["histid_ct_total"] - df2["R1count"])
        )

        df3.cache().createOrReplaceTempView("counted_links")

        print(
            "Round 1 match rate: "
            + str(df3.agg({"R1_pct": "avg"}).collect()[0]["avg(R1_pct)"])
        )
        print(
            "Round 2 match rate of remaining HH members: "
            + str(df3.agg({"R2_pct": "avg"}).collect()[0]["avg(R2_pct)"])
        )

        self.task.spark.sql("set spark.sql.shuffle.partitions=200")



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

import hlink.linking.core.comparison_feature as comparison_feature_core
import hlink.linking.core.comparison as comparison_core

from hlink.linking.link_step import LinkStep


class LinkStepFilter(LinkStep):
    def __init__(self, task):
        super().__init__(
            task,
            "filter",
            input_table_names=["hh_blocked_matches"],
            output_table_names=["hh_potential_matches"],
        )

    def _run(self):
        # self.task.spark.sql("set spark.sql.shuffle.partitions=4000")
        config = self.task.link_run.config

        # establish empty table context dict to pass to SQL template
        t_ctx = {}
        t_ctx["id_col"] = config["id_column"]
        # get comparison_features
        if config.get("hh_comparisons", False):
            t_ctx["matching_clause"] = comparison_core.generate_comparisons(
                config["hh_comparisons"],
                config["comparison_features"],
                config["id_column"],
            )

            comps = comparison_core.get_comparison_leaves(config["hh_comparisons"])
            comp_feature_names = [c["feature_name"] for c in comps]

            t_ctx["feature_columns"] = [
                comparison_feature_core.generate_comparison_feature(
                    f, config["id_column"], include_as=True
                )
                for f in config["comparison_features"]
                if f["alias"] in comp_feature_names
            ]

            self.task.run_register_sql(
                "hh_potential_matches", t_ctx=t_ctx, persist=True
            )

        else:
            self.task.run_register_python(
                "hh_potential_matches",
                lambda: self.task.spark.table("hh_blocked_matches"),
                persist=True,
            )

        self.task.spark.sql("set spark.sql.shuffle.partitions=200")

        print(
            "Potential matches from households which meet hh_comparsions thresholds have been saved to table 'hh_potential_matches'."
        )



# This file is part of the ISRDI's hlink.
# For copyright and licensing information, see the NOTICE and LICENSE files
# in this project's top-level directory, and also on-line at:
#   https://github.com/ipums/hlink

from ..link_task import LinkTask

from .link_step_explode import LinkStepExplode
from .link_step_match import LinkStepMatch
from .link_step_score import LinkStepScore


class Matching(LinkTask):
    def __init__(self, link_run):
        super().__init__(link_run)
        self.training_conf = "training"
        self.table_prefix = ""

    def get_steps(self):
        return [LinkStepExplode(self), LinkStepMatch(self), LinkStepScore(self)]
